package org.zeroturnaround.javarebel.integration.servlet.jasper;

import org.zeroturnaround.bundled.javassist.CannotCompileException;
import org.zeroturnaround.bundled.javassist.ClassPool;
import org.zeroturnaround.bundled.javassist.CtClass;
import org.zeroturnaround.bundled.javassist.CtConstructor;
import org.zeroturnaround.bundled.javassist.CtNewMethod;
import org.zeroturnaround.bundled.javassist.NotFoundException;
import org.zeroturnaround.javarebel.integration.support.JavassistClassBytecodeProcessor;

/**
 * Transforms <code>org.apache.jasper.servlet.JasperLoader</code>
 * by integrating
 * Tomcat (4.x, 5.x, 6.x),
 * JBoss (4.2.x, 5.x),
 * Jetty (5.x) and
 * GlassFish (2.1.x, 3-prelude)
 * JSP pages with JavaRebel.
 * <p>
 * This enables to invoke new methods etc in JSP scriptlets.
 * 
 * @author Rein Raudjärv
 * 
 * @see JspRuntimeContextCBP
 */
public class JasperLoaderCBP extends JavassistClassBytecodeProcessor {

  private static final String REBEL_LOCATER_PREFIX = "com.zeroturnaround.javarebel.RebelLocator$$";
  
  private static final String JSP_PACKAGE_NAME = "org.apache.jsp";
  private static final String JSP_PREFIX = JSP_PACKAGE_NAME.replace('.', '/') + '/';
  private static final int JSP_PREFIX_LENGTH = JSP_PREFIX.length();
  
  public void process(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    if (isTomcat4(cp, cl, ctClass))
      processTomcat4(cp, cl, ctClass);
    else
      registerClassLoader(cp, ctClass);
      
    insertBeforeFindingRebelLocator(cp, cl, ctClass);
  }

  private void registerClassLoader(ClassPool cp, CtClass ctClass) throws CannotCompileException {
    cp.importPackage("org.zeroturnaround.javarebel");
    cp.importPackage("org.zeroturnaround.javarebel.integration.generic");
    
    CtConstructor[] constructors = ctClass.getConstructors();
    for (int i = 0; i < constructors.length; i++) {
        CtConstructor constructor = constructors[i];
        constructor.insertAfter(
            "IntegrationFactory.getInstance().registerClassLoader($0, new RestrictedClassResourceSource($0));");
    }
  }
  
  private boolean isTomcat4(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    try {
      ctClass.getDeclaredMethod("loadClassDataFromFile");
      ctClass.getDeclaredField("className");
      return true;
    } catch (NotFoundException e) {
      return false;
    }
  }
  
  private void processTomcat4(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    cp.importPackage("java.net");
    cp.importPackage("org.apache.jasper");
    cp.importPackage("org.zeroturnaround.javarebel");
    cp.importPackage("org.zeroturnaround.javarebel.integration.generic");
    
    /*
     * Register Class Loader with JavaRebel
     * before the defineClass method is called first time.
     * Otherwise the JSP class would not be instrumented.
     */
    CtConstructor[] constructors = ctClass.getConstructors();
    for (int i = 0; i < constructors.length; i++) {
        CtConstructor constructor = constructors[i];
        constructor.insertAfter(
            "IntegrationFactory.getInstance().registerClassLoader($0, new FindResourceClassResourceSource($0));");
    }
    
    /*
     * If a JavaRebel specific helper class is loaded use the UrlClassLoader.findClass() method
     * to enable JavaRebel to handle the class by itself.
     * 
     * Otherwise the JSP class would be loaded with just a new name. This would be wrong and would cause an error as well.
     */
    ctClass.getDeclaredMethod("loadClass", cp.get(new String[] { "java.lang.String", "boolean"})).insertBefore(
        "{" +
        "  if ($1.startsWith(className) && !$1.equals(className)) {" +
        "    Class clazz = findLoadedClass(name);" +
        "    if (clazz != null) {" +
        "      if ($2)" +
        "        resolveClass(clazz);" +
        "      return clazz;" +
        "    }" +
        
        "    return findClass($1);" +
        "  }" +
        "}");
    
    /*
     * JasperLoader is using the following trick:
     * classes with prefix "org.apache.jsp" are actually loaded from
     * resources without the given prefix.
     * To match the class and resource finding we enable to find resources same way.
     * This enables to load classes without changing the names.
     */
    ctClass.addMethod(CtNewMethod.make(
        "public URL findResource(String name) {" +
        "  if (name.startsWith(\"" + JSP_PREFIX + "\"))" +
        "    name = name.substring(" + JSP_PREFIX_LENGTH + ");" +

        "  return super.findResource(name);;" +
        "}", ctClass));
  }

  private void insertBeforeFindingRebelLocator(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    /*
     * By default JasperLoader loads only classes starting with "org.apache.jsp".
     * So the RebelLocator would not be loaded.
     * 
     * We make the exception and try to find the RebelLocator as well.
     */
    ctClass.getDeclaredMethod("loadClass", cp.get(new String[] { "java.lang.String", "boolean"})).insertBefore(
        "{" +
        "  if ($1.startsWith(\"" + REBEL_LOCATER_PREFIX + "\")) {" +
        "    Class clazz = findLoadedClass($1);" +
        "    if (clazz != null) {" +
        "      if ($2)" +
        "        resolveClass(clazz);" +
        "      return clazz;" +
        "    }" +
        
        "    return super.findClass($1);" +
        "  }" +
        "}");
    
    try {
      ctClass.getDeclaredMethod("findClass").insertBefore(
          "{ synchronized ($0) {" + 
          "    Class result =" + 
          "      findLoadedClass($1);" + 
          "    if (result != null)" + 
          "      return result;" + 
          "    result = " + 
          "      IntegrationFactory.getInstance()" + 
          "        .findReloadableClass($0, $1);" + 
          "    if (result != null)" + 
          "      return result;" + 
          "}}");    
    }
    catch (NotFoundException e) {
      // could not be found
    }
  }  
  
}
