package org.zeroturnaround.javarebel.integration.servlet.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.zeroturnaround.javarebel.ClassEventListener;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;
import org.zeroturnaround.javarebel.Reloader;
import org.zeroturnaround.javarebel.ReloaderFactory;

/**
 * Listener for reinitializing reloaded classes.
 * <p>
 * This is used for classes generated from JSPs to reload constants.
 * 
 * @author Rein Raudjärv
 * 
 * @see Reloader
 */
public class ReinitClassReloadListener implements ClassEventListener {

  private static final Logger log = LoggerFactory.getInstance();
  private static final Reloader reloader = ReloaderFactory.getInstance();

  private static final Set names = new HashSet();
  
  static {
    if (log.isEnabled())
      log.log("Registering " + ReinitClassReloadListener.class.getName());

    reloader.addClassReloadListener(new ReinitClassReloadListener());
  }

  /**
   * Register class to be reinitialized when reloaded.
   * 
   * @param className public class name.
   */
  public static void addClass(String className) {
    if (log.isEnabled())
      log.log("Registering reinitializable class: " + className);

    names.add(className);
  }

  /**
   * Register classes to be reinitialized when reloaded.
   * 
   * @param classNames public class names.
   */
  public static void addClasses(Collection classNames) {
    if (log.isEnabled())
      log.log("Registering reinitializable classes: " + classNames);

    names.addAll(classNames);
  }
  
  // --- ClassEventListener methods ---

  /**
   * Reinitializes the class if it has been registered.
   * 
   * @see #addClass(String)
   * @see #addClasses(Collection)
   */
  public void onClassEvent(int eventType, Class klass) {
    try {
      String name = klass.getName();
      if (names.contains(name)) {
        if (log.isEnabled())
          log.log("Reinitializing class: " + name);

        reloader.reinitClass(klass);
      }
    } catch (Exception e) {
      log.error(e);
    }
  }

  public int priority() {
    return Integer.MAX_VALUE;
  }

}
